# this class represents a single indel as determined from one of many read_indels

class Indel
	attr_reader :read_count, :num_passed_reads, :sample
	def initialize(read_indel, z_score, cutoff, sample=nil)
		if(read_indel.read.dir == :'-')
			@plus_strand_dir = false
			@minus_strand_dir = true
		else
			@plus_strand_dir = true
			@minus_strand_dir = false
		end
		@z_score = z_score
		@indel = read_indel
		@read_count = 1
		@map_qual = read_indel.map_qual
		@ave_nqs = read_indel.avnbq
		@var_rate = read_indel.read.var_rate
		@cutoff = cutoff
		@sample = sample
		if(z_score > cutoff)
			@num_passed_reads = 1
		else
			@num_passed_reads = 0
		end
	end

	# precond: the added read_indel == the current read_indel (but are different reads)
	def  add_read_indel(z_score, read_indel)
		if(read_indel.read.dir == :'-')
			@minus_strand_dir = true
		else
			@plus_strand_dir = true
		end
		@z_score += z_score
		@read_count += 1
		@map_qual += read_indel.map_qual
		@ave_nqs += read_indel.avnbq
		@var_rate += read_indel.read.var_rate
		if(z_score > @cutoff)
			@num_passed_reads += 1
		end
	end

	def z_score
		return ((@z_score.to_f / @read_count.to_f)*10000.0).round/10000.0
	end

	def sum_z_scores
		return (((@z_score.to_f*100.0).round/100.0)*10000.0).round/10000.0
	end

	def read_indel
		return @indel
	end

	def mean_map_qual
		return ((@map_qual.to_f / @read_count.to_f)*100.0).round/100.0
	end


	def mean_avnqs
		return ((@ave_nqs.to_f / @read_count.to_f)*100.0).round/100.0
	end



	def length
		return @indel.length
	end

	def frameshift
		return @indel.frame_shift
	end

	def mean_var_rate
		return ((@var_rate / @read_count.to_f)*100.0).round/100.0
	end

	def simple_local_entropy
		return @indel.simple_local_entropy
	end


	def local_entropy
		return @indel.local_entropy
	end

	# return true if variant reads are only in one strand direction
	def fails_strand_dir_filt
		return !(@plus_strand_dir && @minus_strand_dir)
	end

	def strand_dir
		return 0.0 if fails_strand_dir_filt
		return 1.0
	end

	def site_passed_over_var
		return @num_passed_reads.to_f / @read_count.to_f
	end


	# other must be a read_indel
#	def == (other)
#		if(@indel.is_deletion && other.is_deletion)
#			if(@indel.start_pos == other.start_pos && @indel.ref_seq == other.ref_seq)
#				return true
#			end
#		elsif( ! @indel.is_deletion && ! other.is_deletion)
#			if(@indel.start_pos == other.start_pos && @indel.alt_seq == other.alt_seq)
#				return true
#			end
#		else
#				return false
#		end
#	end


end

