require 'read'
require 'entropy' 
# this abstract class represents a single indel on a single read from a SAM file
class Read_indel
	attr_reader :length, :start_pos, :read

  
	public

	def initialize( read, offset, start_pos, length )
		@read = read
		@offset = offset # how far from the start of the read
		@length = length
		@start_pos = start_pos  # the coordinate of the indel in the genome
		# @pair_end_ratio = read.i_size()/read.seq.size #TODO support multiple indels
		# in a single read
	end


	def map_qual
		return @read.map_qual
	end

	def chr
		return @read.ref
	end


	def read_start_pos
		return @offset
	end


	def frame_shift
		return 0 if @length%3 == 0
		return 1
	end

	def even_length
		return 1 if @length%2 == 0
		return 0
	end


	# # def in_repeat_region? # 	bed = Bed_file.instance() # 	return 1 if
	# bed.pos_included?(@read.ref, @start_pos) # 	return 0 # end


	# returns the distance to the next polymorphism
	def proximity
		cigar = @read.cigar
		cigar_i = nil
		read_pos = 0
		(0...(cigar.length)).each do |i|
			if(read_pos == @offset + 1)
				cigar_i = i
				break
			end
			read_pos += cigar[i][0] unless cigar[i][1] == :"D"
		end

		lower_dist, found_lower_morph = get_morph_dist(cigar, true, cigar_i)
		upper_dist, found_upper_morph = get_morph_dist(cigar, false, cigar_i)

		if(found_lower_morph)
			if(lower_dist < upper_dist || !found_upper_morph)
				return lower_dist
			else
				return upper_dist
			end
		elsif(found_upper_morph)
			return upper_dist
		end

		# did not find any other variations on this read
		return lower_dist if lower_dist > upper_dist
		return upper_dist
	end

	# returns the indel's local entropy, checks 5*indel length on both sides,
	# returns the minimum. Uses indel length as entropy window size
	def local_entropy
		seq_length = (5*@length)
		upper_limit = end_pos + seq_length - 1
		ref = Ref_seq.instance
		if(is_deletion)
			lower_limit = @start_pos - seq_length + 1
			lower_limit = 1 if lower_limit < 1
			lower_seq = ref.get_base_range(chr(), lower_limit, end_pos())
			upper_seq = ref.get_base_range(chr(), @start_pos + 1, upper_limit+1)
		else
			lower_limit = @start_pos - seq_length
			lower_limit = 1 if lower_limit < 1
			lower_seq = "#{ref.get_base_range(chr(), lower_limit, @start_pos-1)}#{alt_seq()}"
			upper_seq = "#{alt_seq()[1..-1]}#{ref.get_base_range(chr(), end_pos(), upper_limit)}"
		end
		upstream_entropy = entropy(lower_seq, @length) 
		dnstream_entropy = entropy(upper_seq, @length)
		if upstream_entropy < dnstream_entropy
			return (upstream_entropy*1000).round/1000.0
		else
			return (dnstream_entropy*1000).round/1000.0
		end
	end


	def simple_local_entropy
		ref = Ref_seq.instance
		lower_limit = @start_pos - 10
		lower_limit += 1 if is_deletion
		lower_limit = 1 if lower_limit < 1
		upper_limit = end_pos() + 10
		return (entropy(ref.get_base_range(chr(), lower_limit, upper_limit), @length) * 1000).round/1000.0
	end

	# returns 1 if the indel is within 5bp of a read end
	def near_read_end
		if(@offset < 4 || @read.seq.length - read_end_pos < 5)
			return 1
		end
		return 0
	end



	# ## abstract methods ## end_pos read_end_pos indel_bases is_deletion

	private

	def get_morph_dist(cigar, is_lower, cigar_i)
		found = false
		dist = 0
		if(is_lower)
			dir = -1
		else
			dir = 1
		end
		entry = cigar[cigar_i+(1*dir)]
		if(entry != nil && entry[1] == :"M")
			ref = Ref_seq.instance
			(0...(entry[0])).each do |i|
				if(is_lower)
					start = @start_pos
					start += 1 if is_deletion
					read_start = read_start_pos
				else
					start = end_pos
					read_start = read_end_pos
				end
				start += (1*dir) if cigar[cigar_i][1] == :"D" #adjust ref_index by 1 for deletions
				# puts "ref[#{start+(i*dir)}] #{ref.get_base(@read.ref, start+(i*dir))}
				# read[#{read_start+(i*dir)}] #{@read.seq[read_start+(i*dir), 1]}"
				if(ref.get_base(@read.ref, start+(i*dir)) != @read.seq[read_start+(i*dir)])
					found = true
					break
				end
				dist += 1
			end
		end
		if(!found && cigar_i+(2*dir) > 0 && cigar_i+(2*dir) < cigar.length)
			found = true
		end
		return dist, found
	end


end #class Indel
